@@ -17,7 +17,7 @@
  */
 
 #include "tclInt.h"
-#include <varargs.h>
+#include <stdarg.h>
 
 /*
  * The following values are used in the flags returned by Tcl_ScanElement
@@ -1010,7 +1010,7 @@ Tcl_SetResult(interp, string, freeProc)
 	/* VARARGS2 */
 #ifndef lint
 void
-Tcl_AppendResult(va_alist)
+Tcl_AppendResult(Tcl_Interp *interp, ...)
 #else
 void
 	/* VARARGS2 */ /* ARGSUSED */
@@ -1020,20 +1020,18 @@ Tcl_AppendResult(interp, p, va_alist)
     char *p;			/* One or more strings to add to the
 				 * result, terminated with NULL. */
 #endif
-    va_dcl
 {
     va_list argList;
-    register Interp *iPtr;
     char *string;
     int newSpace;
-
+    Interp *iPtr = (Interp *) interp;
+    
     /*
      * First, scan through all the arguments to see how much space is
      * needed.
      */
 
-    va_start(argList);
-    iPtr = va_arg(argList, Interp *);
+    va_start(argList,interp);
     newSpace = 0;
     while (1) {
 	string = va_arg(argList, char *);
@@ -1059,8 +1057,7 @@ Tcl_AppendResult(interp, p, va_alist)
      * them into the buffer.
      */
 
-    va_start(argList);
-    (void) va_arg(argList, Tcl_Interp *);
+    va_start(argList,interp);
     while (1) {
 	string = va_arg(argList, char *);
 	if (string == NULL) {
@@ -1256,7 +1253,7 @@ Tcl_ResetResult(interp)
 	/* VARARGS2 */
 #ifndef lint
 void
-Tcl_SetErrorCode(va_alist)
+Tcl_SetErrorCode(Tcl_Interp *interp,...)
 #else
 void
 	/* VARARGS2 */ /* ARGSUSED */
@@ -1266,20 +1263,18 @@ Tcl_SetErrorCode(interp, p, va_alist)
     char *p;			/* One or more elements to add to errorCode,
 				 * terminated with NULL. */
 #endif
-    va_dcl
 {
     va_list argList;
     char *string;
     int flags;
-    Interp *iPtr;
-
+    Interp *iPtr = (Interp *) interp;
+	
     /*
      * Scan through the arguments one at a time, appending them to
      * $errorCode as list elements.
      */
 
-    va_start(argList);
-    iPtr = va_arg(argList, Interp *);
+    va_start(argList,interp);
     flags = TCL_GLOBAL_ONLY | TCL_LIST_ELEMENT;
     while (1) {
 	string = va_arg(argList, char *);
